@@ -17,60 +17,69 @@
  *  nature or kind.
  */
 
-#ifdef OS2
-#define INCL_DOSFILEMGR
-#define INCL_DOSMISC
-#define INCL_DOSQUEUES
-#define INCL_DOSPROCESS
-#include <os2.h>
+#include "bonnie.h"
+
+#include <stdlib.h>
+
+#include "conf.h"
+#ifdef HAVE_ALGORITHM
+#include <algorithm>
 #else
+#ifdef HAVE_ALGO
+#include <algo>
+#else
+#include <algo.h>
+#endif
+#endif
+
 #include <sys/wait.h>
 #include <unistd.h>
-#endif
 #include <sys/time.h>
+#include <pwd.h>
+#include <grp.h>
+#include <sys/utsname.h>
+#include "sync.h"
+
 #include <time.h>
-#include <stdlib.h>
-#include "bonnie.h"
 #include "bon_io.h"
 #include "bon_file.h"
 #include "bon_time.h"
-#include "semaphore.h"
-#include <pwd.h>
-#include <grp.h>
+#include "rand.h"
 #include <ctype.h>
 #include <string.h>
-#include <sys/utsname.h>
 #include <signal.h>
 
-#ifdef AIX_MEM_SIZE
-#include <cf.h>
-#include <sys/cfgodm.h>
-#include <sys/cfgdb.h>
-#endif
-
 void usage();
 
 class CGlobalItems
 {
 public:
   bool quiet;
-  bool fast;
+  int byte_io_size;
   bool sync_bonnie;
+#ifdef O_DIRECT
+  bool use_direct_io;
+#endif
   BonTimer timer;
   int ram;
-  Semaphore sem;
+  Sync *syn;
   char *name;
   bool bufSync;
-  int  chunk_bits;
-  int chunk_size() const { return m_chunk_size; }
+  int  io_chunk_bits;
+  int  file_chunk_bits;
+  int io_chunk_size() const { return m_io_chunk_size; }
+  int file_chunk_size() const { return m_file_chunk_size; }
   bool *doExit;
-  void set_chunk_size(int size)
-    { delete m_buf; m_buf = new char[size]; m_chunk_size = size; }
+  void set_io_chunk_size(int size)
+    { delete m_buf; pa_new(size, m_buf, m_buf_pa); m_io_chunk_size = size; }
+  void set_file_chunk_size(int size)
+    { delete m_buf; m_buf = new char[__max(size, m_io_chunk_size)]; m_file_chunk_size = size; }
 
-  char *buf() { return m_buf; }
+  // Return the page-aligned version of the local buffer
+  char *buf() { return m_buf_pa; }
 
   CGlobalItems(bool *exitFlag);
-  ~CGlobalItems() { delete name; delete m_buf; }
+  ~CGlobalItems() { delete name; delete m_buf; delete syn; }
 
   void decrement_and_wait(int nr_sem);
 
@@ -78,19 +87,32 @@ public:
   {
     delete name;
     name = new char[strlen(path) + 15];
-#ifdef OS2
-    ULONG myPid = 0;
-    DosQuerySysInfo(QSV_FOREGROUND_PROCESS, QSV_FOREGROUND_PROCESS
-                  , &myPid, sizeof(myPid));
-#else
     pid_t myPid = getpid();
-#endif
     sprintf(name, "%s/Bonnie.%d", path, int(myPid));
   }
-private:
-  int m_chunk_size;
-  char *m_buf;
 
+  void setSync(SYNC_TYPE type, int semKey = 0, int num_tests = 0)
+  {
+    syn = new Sync(type, semKey, num_tests);
+  }
+
+private:
+  int m_io_chunk_size;
+  int m_file_chunk_size;
+
+  char *m_buf;     // Pointer to the entire buffer
+  char *m_buf_pa;  // Pointer to the page-aligned version of the same buffer
+
+  // Implement a page-aligned version of new.
+  // 'p' is the pointer created
+  // 'page_aligned_p' is the page-aligned pointer created
+  void pa_new(unsigned int num_bytes, char *&p, char *&page_aligned_p)
+  {
+    int page_size = getpagesize();
+    p = ::new char [num_bytes + page_size];
+
+    page_aligned_p = (char *)((((unsigned long)p + page_size - 1) / page_size) * page_size);
+  }
 
   CGlobalItems(const CGlobalItems &f);
   CGlobalItems & operator =(const CGlobalItems &f);
@@ -98,24 +120,31 @@ private:
 
 CGlobalItems::CGlobalItems(bool *exitFlag)
  : quiet(false)
- , fast(false)
+ , byte_io_size(DefaultByteIO)
  , sync_bonnie(false)
+#ifdef O_DIRECT
+ , use_direct_io(false)
+#endif
  , timer()
  , ram(0)
- , sem(SemKey, TestCount)
+ , syn(NULL)
  , name(NULL)
  , bufSync(false)
- , chunk_bits(DefaultChunkBits)
+ , io_chunk_bits(DefaultChunkBits)
+ , file_chunk_bits(DefaultChunkBits)
  , doExit(exitFlag)
- , m_chunk_size(DefaultChunkSize)
- , m_buf(new char[m_chunk_size])
+ , m_io_chunk_size(DefaultChunkSize)
+ , m_file_chunk_size(DefaultChunkSize)
+ , m_buf(NULL)
+ , m_buf_pa(NULL)
 {
+  pa_new(__max(m_io_chunk_size, m_file_chunk_size), m_buf, m_buf_pa);
   SetName(".");
 }
 
 void CGlobalItems::decrement_and_wait(int nr_sem)
 {
-  if(sem.decrement_and_wait(nr_sem))
+  if(syn->decrement_and_wait(nr_sem))
     exit(1);
 }
 
@@ -126,23 +155,13 @@ int TestFileOps(int file_size, CGlobalItems &globals);
 static bool exitNow;
 static bool already_printed_error;
 
-#ifdef USE_SA_SIGACTION
-#define SIGNAL_NUMBER siginf->si_signo
-#else
-#define SIGNAL_NUMBER sig
-#endif
-
 extern "C"
 {
-  void ctrl_c_handler(int sig
-#ifdef USE_SA_SIGACTION
-		    , siginfo_t *siginf, void *unused
-#endif
-		     )
+  void ctrl_c_handler(int sig, siginfo_t *siginf, void *unused)
   {
-    if(SIGNAL_NUMBER == SIGXCPU)
+    if(siginf->si_signo == SIGXCPU)
       fprintf(stderr, "Exceeded CPU usage.\n");
-    else if(SIGNAL_NUMBER == SIGXFSZ)
+    else if(siginf->si_signo == SIGXFSZ)
       fprintf(stderr, "exceeded file storage limits.\n");
     exitNow = true;
   }
@@ -156,7 +175,7 @@ int main(int argc, char *argv[])
   int    directory_min_size = DefaultDirectoryMinSize;
   int    num_bonnie_procs = 0;
   int    num_directories = 1;
-  int    count = -1;
+  int    test_count = -1;
   const char * machine = NULL;
   char *userName = NULL, *groupName = NULL;
   CGlobalItems globals(&exitNow);
@@ -166,13 +185,8 @@ int main(int argc, char *argv[])
   already_printed_error = false;
 
   struct sigaction sa;
-#ifdef USE_SA_SIGACTION
   sa.sa_sigaction = &ctrl_c_handler;
   sa.sa_flags = SA_RESETHAND | SA_SIGINFO;
-#else
-  sa.sa_handler = ctrl_c_handler;
-  sa.sa_flags = SA_RESETHAND;
-#endif
   if(sigaction(SIGINT, &sa, NULL)
    || sigaction(SIGXCPU, &sa, NULL)
    || sigaction(SIGXFSZ, &sa, NULL))
@@ -180,9 +194,6 @@ int main(int argc, char *argv[])
     printf("Can't handle SIGINT.\n");
     return 1;
   }
-#ifdef USE_SA_SIGACTION
-  sa.sa_sigaction = NULL;
-#endif
   sa.sa_handler = SIG_IGN;
   if(sigaction(SIGHUP, &sa, NULL))
   {
@@ -197,23 +208,19 @@ int main(int argc, char *argv[])
   {
     globals.ram = page_size/1024 * (num_pages/1024);
   }
-#else
-
-#ifdef AIX_MEM_SIZE
-  struct CuAt *odm_obj;
-  int how_many;
-
-  odm_set_path("/etc/objrepos");
-  odm_obj = getattr("sys0", "realmem", 0, &how_many);
-  globals.ram = atoi(odm_obj->value) / 1024;
-  odm_terminate();
-  printf("Memory = %d MB\n", globals.ram);
 #endif
 
-#endif
+  pid_t myPid = 0;
+  myPid = getpid();
+  globals.timer.random_source.seedNum(myPid ^ time(NULL));
+  int concurrency = 1;
 
   int int_c;
-  while(-1 != (int_c = getopt(argc, argv, "bd:fg:m:n:p:qr:s:u:x:y")) )
+  while(-1 != (int_c = getopt(argc, argv, "bc:d:f::g:l:m:n:p:qr:s:u:x:y:z:Z:"
+#ifdef O_DIRECT
+                             "D"
+#endif
+                            )) )
   {
     switch(char(int_c))
     {
@@ -224,6 +231,9 @@ int main(int argc, char *argv[])
       case 'b':
         globals.bufSync = true;
       break;
+      case 'c':
+        concurrency = atoi(optarg);
+      break;
       case 'd':
         if(chdir(optarg))
         {
@@ -232,19 +242,46 @@ int main(int argc, char *argv[])
         }
       break;
       case 'f':
-        globals.fast = true;
+        if(optarg)
+          globals.byte_io_size = atoi(optarg);
+        else
+          globals.byte_io_size = 0;
       break;
       case 'm':
         machine = optarg;
       break;
       case 'n':
-        sscanf(optarg, "%d:%d:%d:%d", &directory_size
-                     , &directory_max_size, &directory_min_size
-                     , &num_directories);
+      {
+        char *sbuf = _strdup(optarg);
+        char *size = strtok(sbuf, ":");
+        directory_size = size_from_str(size, "m");
+        size = strtok(NULL, ":");
+        if(size)
+        {
+          directory_max_size = size_from_str(size, "kmg");
+          size = strtok(NULL, ":");
+          if(size)
+          {
+            directory_min_size = size_from_str(size, "kmg");
+            size = strtok(NULL, ":");
+            if(size)
+            {
+              num_directories = size_from_str(size, "k");
+              size = strtok(NULL, "");
+              if(size)
+              {
+                int tmp = size_from_str(size, "kmg");
+                globals.set_file_chunk_size(tmp);
+              }
+            }
+          }
+        }
+        free(sbuf);
+      }
       break;
       case 'p':
         num_bonnie_procs = atoi(optarg);
-                        /* Set semaphore to # of bonnie++ procs 
+                        /* Set semaphore to # of bonnie++ procs
                            to synchronize */
       break;
       case 'q':
@@ -255,22 +292,21 @@ int main(int argc, char *argv[])
       break;
       case 's':
       {
-        char *sbuf = strdup(optarg);
+        char *sbuf = _strdup(optarg);
         char *size = strtok(sbuf, ":");
-        file_size = atoi(size);
-        char c = size[strlen(size) - 1];
-        if(c == 'g' || c == 'G')
-          file_size *= 1024;
+#ifdef _LARGEFILE64_SOURCE
+        file_size = size_from_str(size, "gt");
+#else
+        file_size = size_from_str(size, "g");
+#endif
         size = strtok(NULL, "");
         if(size)
         {
-          int tmp = atoi(size);
-          c = size[strlen(size) - 1];
-          if(c == 'k' || c == 'K')
-            tmp *= 1024;
-          globals.set_chunk_size(tmp);
+          int tmp = size_from_str(size, "k");
+          globals.set_io_chunk_size(tmp);
         }
         setSize = true;
+        free(sbuf);
       }
       break;
       case 'g':
@@ -282,9 +318,10 @@ int main(int argc, char *argv[])
       {
         if(userName)
           usage();
-        userName = strdup(optarg);
+        userName = _strdup(optarg);
         int i;
-        for(i = 0; userName[i] && userName[i] != ':'; i++);
+        for(i = 0; userName[i] && userName[i] != ':'; i++) {}
+
         if(userName[i] == ':')
         {
           if(groupName)
@@ -295,15 +332,47 @@ int main(int argc, char *argv[])
       }
       break;
       case 'x':
-        count = atoi(optarg);
+        test_count = atoi(optarg);
       break;
       case 'y':
                         /* tell procs to synchronize via previous
                            defined semaphore */
+        switch(tolower(optarg[0]))
+        {
+        case 's':
+          globals.setSync(eSem, SemKey, TestCount);
+        break;
+        case 'p':
+          globals.setSync(ePrompt);
+        break;
+        }
         globals.sync_bonnie = true;
       break;
+      case 'z':
+      {
+        UINT tmp;
+        if(sscanf(optarg, "%u", &tmp) == 1)
+          globals.timer.random_source.seedNum(tmp);
+      }
+      break;
+#ifdef O_DIRECT
+      case 'D':
+        /* open file descriptor with direct I/O */
+        globals.use_direct_io = true;
+      break;
+#endif
+      case 'Z':
+      {
+        if(globals.timer.random_source.seedFile(optarg))
+          return eParam;
+      }
+      break;
     }
   }
+  if(concurrency < 1 || concurrency > 200)
+    usage();
+  if(!globals.syn)
+    globals.setSync(eNone);
   if(optind < argc)
     usage();
 
@@ -315,6 +384,17 @@ int main(int argc, char *argv[])
     if(file_size % 1024 > 512)
       file_size = file_size + 1024 - (file_size % 1024);
   }
+#ifndef _LARGEFILE64_SOURCE
+  if(file_size == 2048)
+    file_size = 2047;
+  if(file_size > 2048)
+  {
+    fprintf(stderr, "Large File Support not present, can't do %dM.\n", file_size);
+    usage();
+  }
+#endif
+  globals.byte_io_size = __min(file_size, globals.byte_io_size);
+  globals.byte_io_size = __max(0, globals.byte_io_size);
 
   if(machine == NULL)
   {
@@ -323,6 +403,9 @@ int main(int argc, char *argv[])
       machine = utsBuf.nodename;
   }
 
+  globals.timer.setMachineName(machine);
+  globals.timer.setConcurrency(concurrency);
+
   if(userName || groupName)
   {
     if(bon_setugid(userName, groupName, globals.quiet))
@@ -341,59 +424,71 @@ int main(int argc, char *argv[])
 
   if(num_bonnie_procs)
   {
+    globals.setSync(eSem, SemKey, TestCount);
     if(num_bonnie_procs == -1)
     {
-      return globals.sem.clear_sem();
+      return globals.syn->clear_sem();
     }
     else
     {
-      return globals.sem.create(num_bonnie_procs);
+      return globals.syn->create(num_bonnie_procs);
     }
   }
 
   if(globals.sync_bonnie)
   {
-    if(globals.sem.get_semid())
+    if(globals.syn->get_semid())
       return 1;
   }
 
   if(file_size < 0 || directory_size < 0 || (!file_size && !directory_size) )
     usage();
-  if(globals.chunk_size() < 256 || globals.chunk_size() > Unit)
+  if(globals.io_chunk_size() < 256 || globals.io_chunk_size() > Unit)
+    usage();
+  if(globals.file_chunk_size() < 256 || globals.file_chunk_size() > Unit)
     usage();
   int i;
-  globals.chunk_bits = 0;
-  for(i = globals.chunk_size(); i > 1; i = i >> 1, globals.chunk_bits++);
-  if(1 << globals.chunk_bits != globals.chunk_size())
+  globals.io_chunk_bits = 0;
+  globals.file_chunk_bits = 0;
+  for(i = globals.io_chunk_size(); i > 1; i = i >> 1, globals.io_chunk_bits++)
+  {}
+
+  if(1 << globals.io_chunk_bits != globals.io_chunk_size())
+    usage();
+  for(i = globals.file_chunk_size(); i > 1; i = i >> 1, globals.file_chunk_bits++)
+  {}
+
+  if(1 << globals.file_chunk_bits != globals.file_chunk_size())
     usage();
 
   if( (directory_max_size != -1 && directory_max_size != -2)
      && (directory_max_size < directory_min_size || directory_max_size < 0
      || directory_min_size < 0) )
     usage();
-  /* If the storage size is too big for the maximum number of files (1000G) */
-  if(file_size > IOFileSize * MaxIOFiles)
+#ifndef _LARGEFILE64_SOURCE
+  if(file_size > (1 << (31 - 20 + globals.io_chunk_bits)) )
+  {
+    fprintf(stderr
+   , "The small chunk size and large IO size make this test impossible in 32bit.\n");
     usage();
-  /* If the file size is so large and the chunk size is so small that we have
-   * more than 2G of chunks */
-  if(globals.chunk_bits < 20 && file_size > (1 << (31 - 20 + globals.chunk_bits)) )
+  }
+#endif
+  if(file_size && globals.ram && (file_size * concurrency) < (globals.ram * 2) )
+  {
+    fprintf(stderr
+          , "File size should be double RAM for good results, RAM is %dM.\n"
+          , globals.ram);
     usage();
+  }
+
   // if doing more than one test run then we print a header before the
   // csv format output.
-  if(count > 1)
+  if(test_count > 1)
   {
     globals.timer.SetType(BonTimer::csv);
     globals.timer.PrintHeader(stdout);
   }
-#ifdef OS2
-    ULONG myPid = 0;
-    DosQuerySysInfo(QSV_FOREGROUND_PROCESS, QSV_FOREGROUND_PROCESS
-                  , &myPid, sizeof(myPid));
-#else
-  pid_t myPid = getpid();
-#endif
-  srand(myPid ^ time(NULL));
-  for(; count > 0 || count == -1; count--)
+  for(; test_count > 0 || test_count == -1; test_count--)
   {
     globals.timer.Initialize();
     int rc;
@@ -404,21 +499,26 @@ int main(int argc, char *argv[])
     if(rc) return rc;
     // if we are only doing one test run then print a plain-text version of
     // the results before printing a csv version.
-    if(count == -1)
+    if(test_count == -1)
     {
       globals.timer.SetType(BonTimer::txt);
-      rc = globals.timer.DoReport(machine, file_size, directory_size
-                                , directory_max_size, directory_min_size
-                                , num_directories, globals.chunk_size()
-                                , globals.quiet ? stderr : stdout);
+      rc = globals.timer.DoReportIO(file_size, globals.byte_io_size
+                    , globals.io_chunk_size(), globals.quiet ? stderr : stdout);
+      rc |= globals.timer.DoReportFile(directory_size
+                    , directory_max_size, directory_min_size, num_directories
+                    , globals.file_chunk_size()
+                    , globals.quiet ? stderr : stdout);
     }
     // print a csv version in every case
     globals.timer.SetType(BonTimer::csv);
-    rc = globals.timer.DoReport(machine, file_size, directory_size
-                              , directory_max_size, directory_min_size
-                              , num_directories, globals.chunk_size(), stdout);
+    rc = globals.timer.DoReportIO(file_size, globals.byte_io_size
+                   , globals.io_chunk_size(), stdout);
+    rc |= globals.timer.DoReportFile(directory_size
+                    , directory_max_size, directory_min_size, num_directories
+                    , globals.file_chunk_size(), stdout);
     if(rc) return rc;
   }
+  return eNoErr;
 }
 
 int
@@ -426,73 +526,83 @@ TestFileOps(int file_size, CGlobalItems &globals)
 {
   if(file_size)
   {
-    CFileOp file(globals.timer, file_size, globals.chunk_bits, globals.bufSync);
+    CFileOp file(globals.timer, file_size, globals.io_chunk_bits, globals.bufSync
+#ifdef O_DIRECT
+               , globals.use_direct_io
+#endif
+               );
     int    num_chunks;
     int    words;
     char  *buf = globals.buf();
     int    bufindex;
     int    i;
 
-    if(globals.ram && file_size < globals.ram * 2)
-    {
-      fprintf(stderr
-            , "File size should be double RAM for good results, RAM is %dM.\n"
-            , globals.ram);
-      return 1;
-    }
-    // default is we have 1M / 8K * 200 chunks = 25600
-    num_chunks = Unit / globals.chunk_size() * file_size;
+    // default is we have 1M / 8K * 300 chunks = 38400
+    num_chunks = Unit / globals.io_chunk_size() * file_size;
+    int char_io_chunks = Unit / globals.io_chunk_size() * globals.byte_io_size;
 
     int rc;
-    rc = file.open(globals.name, true, true);
+    rc = file.Open(globals.name, true);
     if(rc)
       return rc;
     if(exitNow)
-      return EXIT_CTRL_C;
-    globals.timer.timestamp();
+      return eCtrl_C;
+    Duration dur;
 
-    if(!globals.fast)
+    globals.timer.start();
+    if(char_io_chunks)
     {
-      globals.decrement_and_wait(Putc);
-      // Fill up a file, writing it a char at a time with the stdio putc() call
-      if(!globals.quiet) fprintf(stderr, "Writing with putc()...");
-      for(words = 0; words < num_chunks; words++)
+      dur.reset();
+      globals.decrement_and_wait(ByteWrite);
+      // Fill up a file, writing it a char at a time
+      if(!globals.quiet) fprintf(stderr, "Writing a byte at a time...");
+      for(words = 0; words < char_io_chunks; words++)
       {
-        if(file.write_block_putc() == -1)
+        dur.start();
+        if(file.write_block_byte() == -1)
           return 1;
+        dur.stop();
         if(exitNow)
-          return EXIT_CTRL_C;
+          return eCtrl_C;
       }
       fflush(NULL);
       /*
        * note that we always close the file before measuring time, in an
        *  effort to force as much of the I/O out as we can
        */
-      file.close();
-      globals.timer.get_delta_t(Putc);
+      file.Close();
+      globals.timer.stop_and_record(ByteWrite);
+      globals.timer.add_latency(ByteWrite, dur.getMax());
       if(!globals.quiet) fprintf(stderr, "done\n");
     }
     /* Write the whole file from scratch, again, with block I/O */
     if(file.reopen(true))
       return 1;
+    dur.reset();
     globals.decrement_and_wait(FastWrite);
     if(!globals.quiet) fprintf(stderr, "Writing intelligently...");
-    memset(buf, 0, globals.chunk_size());
-    globals.timer.timestamp();
+    memset(buf, 0, globals.io_chunk_size());
+    globals.timer.start();
     bufindex = 0;
     // for the number of chunks of file data
     for(i = 0; i < num_chunks; i++)
     {
       if(exitNow)
-        return EXIT_CTRL_C;
+        return eCtrl_C;
       // for each chunk in the Unit
       buf[bufindex]++;
-      bufindex = (bufindex + 1) % globals.chunk_size();
+      bufindex = (bufindex + 1) % globals.io_chunk_size();
+      dur.start();
       if(file.write_block(PVOID(buf)) == -1)
-        return io_error("write(2)");
+      {
+        fprintf(stderr, "Can't write block %d.\n", i);
+        return 1;
+      }
+      dur.stop();
     }
-    file.close();
-    globals.timer.get_delta_t(FastWrite);
+    file.Close();
+    globals.timer.stop_and_record(FastWrite);
+    globals.timer.add_latency(FastWrite, dur.getMax());
     if(!globals.quiet) fprintf(stderr, "done\n");
 
 
@@ -504,48 +614,55 @@ TestFileOps(int file_size, CGlobalItems &globals)
       if(!globals.quiet) fprintf(stderr, "error in lseek(2) before rewrite\n");
       return 1;
     }
+    dur.reset();
     globals.decrement_and_wait(ReWrite);
     if(!globals.quiet) fprintf(stderr, "Rewriting...");
-    globals.timer.timestamp();
+    globals.timer.start();
     bufindex = 0;
     for(words = 0; words < num_chunks; words++)
     { // for each chunk in the file
+      dur.start();
       if (file.read_block(PVOID(buf)) == -1)
         return 1;
-      bufindex = bufindex % globals.chunk_size();
+      bufindex = bufindex % globals.io_chunk_size();
       buf[bufindex]++;
       bufindex++;
       if (file.seek(-1, SEEK_CUR) == -1)
         return 1;
       if (file.write_block(PVOID(buf)) == -1)
         return io_error("re write(2)");
+      dur.stop();
       if(exitNow)
-        return EXIT_CTRL_C;
+        return eCtrl_C;
     }
-    file.close();
-    globals.timer.get_delta_t(ReWrite);
+    file.Close();
+    globals.timer.stop_and_record(ReWrite);
+    globals.timer.add_latency(ReWrite, dur.getMax());
     if(!globals.quiet) fprintf(stderr, "done\n");
 
-
-    if(!globals.fast)
+    if(char_io_chunks)
     {
-      // read them all back with getc()
-      if(file.reopen(false, true))
+      // read them all back a byte at a time
+      if(file.reopen(false))
         return 1;
-      globals.decrement_and_wait(Getc);
-      if(!globals.quiet) fprintf(stderr, "Reading with getc()...");
-      globals.timer.timestamp();
+      dur.reset();
+      globals.decrement_and_wait(ByteRead);
+      if(!globals.quiet) fprintf(stderr, "Reading a byte at a time...");
+      globals.timer.start();
 
-      for(words = 0; words < num_chunks; words++)
+      for(words = 0; words < char_io_chunks; words++)
       {
-        if(file.read_block_getc(buf) == -1)
+        dur.start();
+        if(file.read_block_byte(buf) == -1)
           return 1;
+        dur.stop();
         if(exitNow)
-          return EXIT_CTRL_C;
+          return eCtrl_C;
       }
 
-      file.close();
-      globals.timer.get_delta_t(Getc);
+      file.Close();
+      globals.timer.stop_and_record(ByteRead);
+      globals.timer.add_latency(ByteRead, dur.getMax());
       if(!globals.quiet) fprintf(stderr, "done\n");
     }
 
@@ -554,22 +671,26 @@ TestFileOps(int file_size, CGlobalItems &globals)
       return 1;
     if (file.seek(0, SEEK_SET) == -1)
       return io_error("lseek before read");
+    dur.reset();
     globals.decrement_and_wait(FastRead);
     if(!globals.quiet) fprintf(stderr, "Reading intelligently...");
-    globals.timer.timestamp();
+    globals.timer.start();
     for(i = 0; i < num_chunks; i++)
     { /* per block */
+      dur.start();
       if ((words = file.read_block(PVOID(buf))) == -1)
         return io_error("read(2)");
+      dur.stop();
       if(exitNow)
-        return EXIT_CTRL_C;
+        return eCtrl_C;
     } /* per block */
-    file.close();
-    globals.timer.get_delta_t(FastRead);
+    file.Close();
+    globals.timer.stop_and_record(FastRead);
+    globals.timer.add_latency(FastRead, dur.getMax());
     if(!globals.quiet) fprintf(stderr, "done\n");
 
-    globals.timer.timestamp();
-    if(file.seek_test(globals.quiet, globals.sem))
+    globals.timer.start();
+    if(file.seek_test(globals.timer.random_source, globals.quiet, *globals.syn))
       return 1;
 
     /*
@@ -590,14 +711,14 @@ TestFileOps(int file_size, CGlobalItems &globals)
      *  all the children are lseeking on the same underlying file object?
      */
   }
-  return 0;
+  return eNoErr;
 }
 
 int
 TestDirOps(int directory_size, int max_size, int min_size
          , int num_directories, CGlobalItems &globals)
 {
-  COpenTest open_test(globals.chunk_size(), globals.bufSync, globals.doExit);
+  COpenTest open_test(globals.file_chunk_size(), globals.bufSync, globals.doExit);
   if(!directory_size)
   {
     return 0;
@@ -609,18 +730,18 @@ TestDirOps(int directory_size, int max_size, int min_size
   if(globals.ram && directory_size * MaxDataPerFile * 2 > (globals.ram << 10))
   {
     fprintf(stderr
-         , "When testing %dK of files in %d MB of RAM the system is likely to\n"
+        , "When testing %dK of files in %d MiB of RAM the system is likely to\n"
            "start paging Bonnie++ data and the test will give suspect\n"
            "results, use less files or install more RAM for this test.\n"
           , directory_size, globals.ram);
-    return 1;
+    return eParam;
   }
   // Can't use more than 1G of RAM
   if(directory_size * MaxDataPerFile > (1 << 20))
   {
     fprintf(stderr, "Not enough ram to test with %dK files.\n"
                   , directory_size);
-    return 1;
+    return eParam;
   }
   globals.decrement_and_wait(CreateSeq);
   if(!globals.quiet) fprintf(stderr, "Create files in sequential order...");
@@ -651,21 +772,23 @@ TestDirOps(int directory_size, int max_size, int min_size
   if(open_test.delete_random(globals.timer))
     return 1;
   if(!globals.quiet) fprintf(stderr, "done.\n");
-  return 0;
+  return eNoErr;
 }
 
 void
 usage()
 {
-  fprintf(stderr,
-    "usage: bonnie++ [-d scratch-dir] [-s size(Mb)[:chunk-size(b)]]\n"
-    "                [-n number-to-stat[:max-size[:min-size][:num-directories]]]\n"
-    "                [-m machine-name]\n"
-    "                [-r ram-size-in-Mb]\n"
-    "                [-x number-of-tests] [-u uid-to-use:gid-to-use] [-g gid-to-use]\n"
-    "                [-q] [-f] [-b] [-p processes | -y]\n"
+  fprintf(stderr, "usage:\n"
+    "bonnie++ [-d scratch-dir] [-c concurrency] [-s size(MiB)[:chunk-size(b)]]\n"
+    "      [-n number-to-stat[:max-size[:min-size][:num-directories[:chunk-size]]]]\n"
+    "      [-m machine-name] [-r ram-size-in-MiB]\n"
+    "      [-x number-of-tests] [-u uid-to-use:gid-to-use] [-g gid-to-use]\n"
+    "      [-q] [-f] [-b] [-p processes | -y] [-z seed | -Z random-file]\n"
+#ifdef O_DIRECT
+    "      [-D]\n"
+#endif
     "\nVersion: " BON_VERSION "\n");
-  exit(1);
+  exit(eParam);
 }
 
 int
